(* Andy's trees.ml file.  
   Basically the same as Dan's except more verbose identifiers;
   where our solutions differed, I mixed and matched some of
   the styles *)

(* This implements a set interface: multiple insertions of the
   same element have no effect *)

(* -------------------------------------------------------------- *)
(* Integer BST *)

type intbst = IntEmpty | IntNode of int * intbst * intbst


let rec int_insert tree element =
   match tree with
      IntEmpty -> IntNode(element, IntEmpty, IntEmpty)
    | IntNode(nodeElement, left, right) ->
         if element = nodeElement then 
            IntNode(nodeElement, left, right)
         else if element < nodeElement then
            IntNode(nodeElement, int_insert left element, right)
         else
            IntNode(nodeElement, left, int_insert right element)


let rec int_lookup tree element = 
   match tree with
      IntEmpty -> false
    | IntNode(nodeElement, left, right) ->
         if element = nodeElement then 
            true
         else if element < nodeElement then
            int_lookup left element
         else
            int_lookup right element


let rec int_sum tree =
   match tree with
      IntEmpty -> 0
    | IntNode(nodeElement, left, right) ->
         nodeElement + int_sum left + int_sum right


exception NoElementGreaterThan

(* This is the least_greater solution that 
   catches its own exception sometimes *)

let rec int_least_greater tree value = 
   match tree with
      IntEmpty -> raise NoElementGreaterThan
    | IntNode(nodeElement, left, right) ->
         if value >= nodeElement then
            int_least_greater right value
         else try
            int_least_greater left value
         with NoElementGreaterThan ->
            nodeElement


(* -------------------------------------------------------------- *)
(* Polymorpic BST *)

(* An 'a bst is just a "bare" BST that holds 'a s *)
type 'a bst = Empty | Node of 'a * 'a bst * 'a bst

(* The user-supplied comparison function returns one of these *)
type comparison_result = LessThan | GreaterThan | Equal

(* The whole data structure is a "bare" bst and a comparison function *)
type 'a bst_and_comp = 'a bst * ('a -> 'a -> comparison_result)


(* To handle the distinction between the "bare" BST and the
   whole data structure, each whole data structure function
   uses a recursive bare BST helper function *)


(* This one uses the arguably clumsy approach of passing
   the element as a parameter to insert_rec *)

let insert wholeTree element = 
   let bareTree,comparisonFunc = wholeTree in
   let rec insert_rec tree elt =
      match tree with
         Empty -> Node(elt, Empty, Empty)
       | Node(nodeElement, left, right) ->
            match comparisonFunc elt nodeElement with
               Equal       -> Node(nodeElement, left, right)
             | LessThan    -> Node(nodeElement, insert_rec left elt, right)
             | GreaterThan -> Node(nodeElement, left, insert_rec right elt)
   in insert_rec bareTree element


(* This one uses the arguably more element approach of 
   not passing the element as an argument because it 
   is already bound in the context of lookup_rec *)

let lookup wholeTree element = 
   let bareTree,comparisonFunc = wholeTree in
   let rec lookup_rec tree = 
      match tree with
         Empty -> false
       | Node(nodeElement, left, right) ->
            match comparisonFunc element nodeElement with
               Equal       -> true
             | LessThan    -> lookup_rec left
             | GreaterThan -> lookup_rec right
   in lookup_rec bareTree


(* This also uses the "not-passing-element" approach (although
   here the argument is named "value" rather than "element"), 
   and in addition illustrates the technique of using the 
   predefined option type within the helper, and only throwing 
   an exception in the outer function *)

let least_greater wholeTree value =
   let bareTree,comparisonFunc = wholeTree in
   let rec least_greater_rec tree =
      match tree with
         Empty -> None
       | Node(nodeElement, left, right) ->
            match comparisonFunc value nodeElement with
               Equal | GreaterThan -> least_greater_rec right
             | LessThan ->
                  let leftAnswer = least_greater_rec left in
                  match leftAnswer with
                     Some x -> leftAnswer
                   | None -> Some nodeElement
   in let result = least_greater_rec bareTree in
   match result with
      Some x -> x
    | None -> raise NoElementGreaterThan

(* Interestingly, because lookup and least_greater never compare 
   elements within the tree, and instead always compare their
   argument with one element at a time, they have a slightly 
   more general type than one might otherwise expect, because
   the type of the argument does not have to match the type 
   in the tree as long as the comparison function can connect
   them *)
